"""Kitt configuration utils"""

import os
import json
import toml

from kitt.logger import panic, debug


class ConfigUtils:
    """ConfigUtils

    Kitt configuration utils.
    """

    @staticmethod
    def mkpath(relative: str) -> str:
        """Create absolute path from relative.

        Args:
            relative (str): relative path to ressource

        Returns:
            str: absolute path using app current folder
        """
        local = os.path.dirname(__file__)
        return os.path.join(local, relative)

    @classmethod
    def load(cls, config_file: str = None) -> dict:
        """Load config from file, override default config

        Args:
            config_file (str, optional): configuration file path. Defaults to None.

        Returns:
            dict: loaded configuration
        """
        custom = {}
        default = None
        default_config_path = cls.mkpath('static/default.toml')
        try:
            if not (default := toml.load(default_config_path)):
                raise toml.TomlDecodeError
        except (toml.TomlDecodeError, FileNotFoundError, IOError):
            panic('Could not load default config file.')

        if not config_file:
            return default

        try:
            if not (custom := toml.load(config_file)):
                raise ValueError

        except (FileNotFoundError, IOError):
            panic('File does not exist or is not readable.')

        except toml.TomlDecodeError:
            if not (custom := json.load(config_file)):
                raise ValueError from toml.TomlDecodeError

        except ValueError:
            panic('File content cannot be properly handled.')

        except Exception as exception:
            debug(exception)
            panic('Error loading config files.')

        # Handle incomplete user config file
        default = cls.update(custom, default)

        return default

    @classmethod
    def update(cls, src: dict, dest: dict) -> dict:
        """
        Deep update dict config `dest' with config `src'
        Make sure it wont overwrite if empty str or list.

        Args:
            src (dict): config to update
            dest (dict): config override

        Returns:
            dict: updated configuration
        """
        for key, value in src.items():
            if key in dest and isinstance(value, dict) and isinstance(dest[key], dict):
                dest[key] = cls.update(src[key], dest[key])
            elif key in dest and value in ['', [], ['']]:
                continue
            else:
                dest[key] = src[key]

        return dest



"""Logging and debugging utilities"""

import traceback

from typing import Union

from rich.console import Console

console = Console()
console.debug = False


def success(msg: str):
    """Display cli success

    Args:
        msg (str): success message
    """
    console.print('[green]✓ ' + msg)


def info(msg: str):
    """Display cli info

    Args:
        msg (str): info message
    """
    console.print('[sky_blue3] ' + msg)


def warning(msg: str):
    """Display cli warning

    Args:
        msg (str): warning message
    """
    console.print('[yellow]~ ' + msg)


def waiter(msg: str):
    """Provide pending status object

    Args:
        msg (str): waiter message
    """
    return console.status('[bold grey]' + msg)


def debug(msg: Union[str, Exception]):
    """Extended log for debug mode

    Args:
        msg (Union[str, Exception]): debug data to log
    """
    if console.debug:
        if not isinstance(msg, str):
            console.log('[grey62]' + str(traceback.format_exc()))
        console.log('[grey62]' + str(msg))


def config(debug_mode: bool = False):
    """Config logger

    Args:
        debug_mode (bool, optional): set debug mode. Defaults to False.
    """
    console.debug = debug_mode


def panic(msg: str):
    """Program panic

    Args:
        msg (str): panic message
    """
    console.print('[red]✗ ' + msg)
    exit(1)



"""Handles basic crypto and kitt vault managment"""

import json
import base64
import getpass
from hashlib import sha256
from cryptography.fernet import Fernet


def _forge_key(data: str) -> bytes:
    try:
        _hash = sha256(data.encode('utf8'))
        key = _hash.digest()
        return base64.urlsafe_b64encode(key)
    except Exception:
        return None


def secure_prompt() -> str:
    """Show secure hidden pasword prompt

    Returns:
        str: data
    """
    try:
        return getpass.getpass()
    except getpass.GetPassWarning:
        return None

def cipher_dict(password: str, vault: dict) -> str:
    """Cipher kitt vault

    Args:
        password (str): password as encryption key
        vault (dict): secrets

    Returns:
        str: base64 encoded encrypted vault
    """

    if not isinstance(vault, dict):
        raise ValueError

    try:
        vault = json.dumps(vault)
    except TypeError:
        return None

    return cipher_text(password, vault)


def uncipher_dict(password: str, vault: str) -> dict:
    """Uncipher kitt vault

    Args:
        password (str): password as decryption key
        vault (str): encrypted vault as string

    Returns:
        dict: kitt vault as secret list
    """

    vault = uncipher_text(password, vault)
    
    try:
        vault = json.loads(vault)
        if not isinstance(vault, dict):
            raise ValueError
        return vault
    except (ValueError, TypeError):
        return None


def cipher_text(password: str, data: str) -> str:
    """Simple text encryption

    Args:
        password (str): password as encryption key
        data (str): data to cipher

    Returns:
        str: encrypted data as string
    """
    
    if not (key := _forge_key(password)):
        raise ValueError
    
    try:
        cipher = Fernet(key).encrypt(data.encode('utf-8'))
        return base64.b64encode(cipher).decode('utf-8')
    except Exception:
        return None


def uncipher_text(password: str, data: str) -> str:
    """Simple text decryption

    Args:
        password (str): password as decryption key
        data (str): data to decipher

    Returns:
        str: decrypted data as string
    """
    key = _forge_key(password)
    try:
        if not key:
            raise ValueError
        dcode = base64.b64decode(data)
        return Fernet(key).decrypt(dcode).decode('utf-8')
    except Exception:
        return None


def b64(data: bytes) -> str:
    """Base64 encode

    Args:
        data (bytes): data to encode

    Returns:
        str: encoded data
    """
    try:
        return base64.b64encode(data).decode('utf-8')
    except Exception:
        return None


def b64d(data: str) -> bytes:
    """Base64 decode

    Args:
        data (str): encoded data

    Returns:
        bytes: decoded data
    """
    try:
        return base64.b64decode(data.encode('utf-8'))
    except Exception:
        return None



"""Provides low-level container image abstractions"""

import io

from abc import ABC as AbstractClass, abstractmethod

import jinja2
import docker
import dockerpty


class Composer:
    """Image text file composer"""

    def __init__(self, template: str):
        """Init

        Args:
            template (str): jinja2 template
        """

        with open(template, 'r', encoding='utf-8') as file:
            self.template = file.read()

    def compose(self, values: dict) -> str:
        """Copose file from template

        Args:
            values (dict): templating values

        Returns:
            str: generated templated text file
        """

        template = jinja2.Template(self.template)
        return template.render(values)


class ImageManager(AbstractClass):
    """Abstract ImageManager

    Should be implemented by any image manager (docker, podman, ..)
    """

    @abstractmethod
    def run(self, name: str, tag: str = 'latest', **kwargs):
        """Create container, run and attach to current TTY.

        Args:
            name (str): image to run.
            tag (str, optional): image tag. Defaults to 'latest'.
            **kwargs (Any): any argument that undelying run method would accept.
        """
        raise NotImplementedError

    @abstractmethod
    def list(self, repository: str = None):
        """List local images

        Args:
            repository (str, optional): only for this repository. Defaults to None.
        """
        raise NotImplementedError

    @abstractmethod
    def remove(self, name: str, tag: str = 'latest'):
        """Remove local image

        Args:
            name (str): image name.
            tag (str, optional): image tag. Defaults to 'latest'.
        """
        raise NotImplementedError

    @abstractmethod
    def prune(self, repository: str = None):
        """Prune local images

        Args:
            repository (str, optional): only for this repository. Defaults to None.
        """
        raise NotImplementedError

    @abstractmethod
    def refresh(self, repository: str = None):
        """Pull latest version of local images

        Args:
            repository (str, optional): only for this repository. Defaults to None.
        """
        raise NotImplementedError

    @abstractmethod
    def build(self, name: str, template: str, tag: str = 'latest', **kwargs):
        """Build image

        Args:
            name (str): image name.
            template (str): text object to build image from (ex. Dockerfile).
            tag (str, optional): image tag. Defaults to 'latest'.
            **kwargs (Any): any argument that undelying build method would accept.
        """
        raise NotImplementedError

    @abstractmethod
    def pull(self, repository: str, tag: str = 'latest'):
        """Pull image from registry

        Args:
            repository (str): full repository url.
            tag (str, optional): image tag. Defaults to 'latest'.
            alias (str, optional): image name repository alias. Defaults to 'latest'.
        """
        raise NotImplementedError

    @abstractmethod
    def push(self, repository: str, name: str, tag: str = 'latest', clean: bool = True):
        """Push image to remote registry

        Args:
            repository (str): full repository url.
            name (str): local image name to push.
            tag (str, optional): image tag. Defaults to 'latest'.
            clean (bool, optional): Remove generated local tag. Defaults to True.
        """
        raise NotImplementedError

    @abstractmethod
    def stat(self, name: str, tag: str = 'latest'):
        """Stat if local image exists

        Args:
            name (str): image name.
            tag (str, optional): image tag. Defaults to 'latest'.
        """
        raise NotImplementedError

    @abstractmethod
    def labels(self, name: str, tag: str = 'latest') -> dict:
        """Get local image labels

        Args:
            name (str): image name.
            tag (str, optional): image tag. Defaults to 'latest'.

        Returns:
            dict: labels as dictionary.
        """
        raise NotImplementedError


def docker_error_handler(func):
    """Docker Error Generic Handler

    Args:
        func (types.FunctionType): _description_
    """

    def _handler(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except docker.errors.BuildError as error:
            self.logger.debug(error)
            self.logger.panic('Docker build error')
        except docker.errors.ImageNotFound as error:
            self.logger.debug(error)
            self.logger.panic('Local docker image not found')
        except docker.errors.APIError as error:
            self.logger.debug(error)
            self.logger.panic('Docker API error')
        except docker.errors.DockerException as error:
            self.logger.debug(error)
            self.logger.panic('Problem with docker daemon (--debug)')

    return _handler

# ---------------+
#     Docker     |
# ---------------+


class DockerImageManager(ImageManager):
    """Docker image manager (child class of ImageManager)"""

    @staticmethod
    def _tag(name, tag):
        return f'{ name }:{ tag }'

    @docker_error_handler
    def __init__(self, logger):
        self.logger = logger
        self.client = docker.from_env()
        self.experimental = self.client.info().get('ExperimentalBuild', False)

    @docker_error_handler
    def _get(self, name: str, tag: str = 'latest'):
        fname = self._tag(name, tag)
        return self.client.images.get(fname)

    @docker_error_handler
    def _get_all(self, repository: str = None) -> [docker.models.images.Image]:
        return self.client.images.list(name=repository)

    @docker_error_handler
    def run(self, name: str, tag: str = 'latest', **kwargs):
        fname = self._tag(name, tag)
        container = self.client.containers.create(
            image=fname,
            auto_remove=True,
            stdin_open=True,
            tty=True,
            detach=False,
            network_mode='host',
            **kwargs
        )
        dockerpty.start(self.client.api, container.id)

    @docker_error_handler
    def list(self, repository: str = None):
        images = self._get_all(repository)
        for image in images:
            basename = [x[len(repository) + 1:]
                        for x in image.tags if x.startswith(repository)].pop()
            self.logger.info('➜ ' + basename)

    @docker_error_handler
    def remove(self, name: str, tag: str = 'latest'):
        self.client.images.remove(self._tag(name, tag))

    @docker_error_handler
    def prune(self, repository: str = None):
        self.client.images.prune()
        if not repository:
            return
        for image in self._get_all(repository):
            for tag in image.tags:
                if tag.startswith(repository):
                    self.remove(*tag.split(':'))

    @docker_error_handler
    def refresh(self, repository: str = None):
        images = self._get_all(repository)
        map(lambda x: x.reload(), images)

    @docker_error_handler
    def build(self, name: str, template: str, tag: str = 'latest', squash = True, **kwargs):
        dockerfile = io.BytesIO(template.encode('utf-8'))
        fname = self._tag(name, tag)
        self.client.images.build(
            tag=fname,
            fileobj=dockerfile,
            rm=True,
            nocache=True,
            squash=squash and self.experimental,
            **kwargs
        )

    @docker_error_handler
    def pull(self, repository: str, tag: str = 'latest', alias: str = None):
        uri = f'{ repository }:{ tag }'
        if not (image := self.client.images.pull(uri)):
            self.logger.panic(f'Nothing to pull at { uri }')
        if alias and isinstance(alias, str):
            image.tag(alias, tag)

    @docker_error_handler
    def push(self, repository: str, name: str, tag: str = 'latest', clean: bool = True):
        self.logger.warning('Push status is not perfectly handled yet.')
        self.logger.warning('To see full push trace, run with --debug option.')

        image = self._get(name, tag)
        image.tag(repository, tag)
        output = self.client.images.push(repository, tag)
        self.logger.debug(output)
        if clean:
            self.remove(repository, tag)

    @docker_error_handler
    def stat(self, name: str, tag: str = 'latest') -> bool:
        return self._get(name, tag)

    @docker_error_handler
    def labels(self, name: str, tag: str = 'latest') -> dict:
        image = self._get(name, tag)
        return image.labels if image else {}



"""Kitt commands handler - Core module"""

import os
import grp
import json
import tempfile
import subprocess

from typing import Tuple
from pwd import getpwnam

from kitt import plugins
from kitt.__version__ import __version__
from kitt.config import ConfigUtils
from kitt.images import DockerImageManager, Composer
from kitt.vault import create_vault, load_vault, VaultFS
from kitt import logger
from kitt.logger import (
    success,
    info,
    warning,
    waiter,
    panic,
)


class KittClient:
    """Kitt Client"""

    def __init__(self):
        template = ConfigUtils.mkpath('static/Dockerfile.j2')
        self.image_composer = Composer(template)
        self.image_manager = DockerImageManager(logger)

    def run(self, name: str, extras: dict = None):
        """Run kitt shell

        Args:
            name (str): kitt image specifier
            extras (dict, optional): extras runtime configurations. Defaults to None.
        """
        if not self.image_manager.stat('kitt', name):
            panic(f'Image { name } not found. Use `pull` command first.')

        envs = {}
        volumes = {}
        groups = []

        user_uid = os.getuid()
        user_gid = os.getuid()
        user_home = os.environ.get('HOME')

        extras = extras or {}

        if username := extras.get('run_as'):
            try:
                user = getpwnam(username)
                user_uid = user.pw_uid
                user_gid = user.pw_gid
            except KeyError:
                panic(f'Could not get "{ username }" user infos')

        config = self._config(name)
        hostname = config.get('hostname', 'kitt')

        if config.get('dind') or extras.get('dind'):
            try:
                host_docker_gid = grp.getgrnam('docker').gr_gid
                groups.append(host_docker_gid)
            except KeyError:
                warning('Could not find host group "docker"')

        volumes = {**volumes, **config.get('bind_volumes', {})}
        for vol in dict(volumes):
            if user_home and '$HOME' in vol:
                host = vol.replace('$HOME', user_home)
                volumes[host] = volumes.pop(vol)

        for vol in extras.get('volumes', []):
            host, bind, mode = unpack_volume(vol)
            volumes[host] = {'bind': bind, 'mode': mode}

        vault_fs = None
        while str_vault := config.get('vault'):
            vault = load_vault(str_vault)
            if not vault:
                warning('Invalid password or corrupted vault')
                break
            envs = {**envs, **vault.get('envs', {})}
            if files := vault.get('files'):
                vault_fs = VaultFS()
                volumes_fs = vault_fs.load(files)
                volumes = {**volumes, **volumes_fs}
            break

        # As container network is in host mode, will exploit Xorg
        # abstract socket instead of /tmp/.X11-unix socket
        if config.get('forward_x11'):
            envs['DISPLAY'] = os.environ.get('DISPLAY')
            # Is run as root, probably not in xhost auth list
            if user_uid == 0:
                subprocess.call(
                    args=f'xhost +local:{ hostname }',
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.STDOUT
                )

        self.image_manager.run(
            name='kitt',
            tag=name,
            hostname=hostname,
            volumes=volumes,
            environment=envs,
            cap_add=['CAP_NET_RAW', 'CAP_NET_ADMIN', 'CAP_IPC_LOCK'],
            extra_hosts={hostname: '127.0.0.1'},
            command=config.get('command', 'bash'),
            group_add=groups,
            user=f'{ user_uid }:{ user_gid }',
        )

        if vault_fs:
            vault_fs.close()

    def build(self, name: str, config_file: str):
        """Build kitt image using provided config file

        Args:
            name (str): kitt image name
            config_file (str): config file path
            catalog (str): custom catalog file
        """

        config = ConfigUtils.load(config_file)
        workspace = config.get('workspace')
        options = config.get('options')

        context = {
            'user': workspace.get('user', 'user'),
            'shell': workspace.get('default_shell', 'bash'),
            'tools': workspace.get('tools', []),
            'envs': workspace.get('envs', []),
            'paths': workspace.get('paths', []),
            'image': workspace.get('image', 'ubuntu:22.04'),
            'plugins': [plugins.compose(n, c) for n, c in config.get('plugins', {}).items()],
        }

        template = self.image_composer.compose(context)
        volumes = {}

        if options.get('docker_in_docker'):
            sock = '/var/run/docker.sock'
            volumes[sock] = {'bind': sock, 'mode': 'rw'}

        for vol in workspace.get('volumes', []):
            host = vol.get('host')
            bind = vol.get('bind')
            mode = vol.get('mode', 'rw')

            if not host or not bind:
                warning(f'Bad volume format : "{vol}"')
                continue

            volumes[host] = {'bind': bind, 'mode': mode}

        secrets = config.get('secrets', {})
        str_vault = create_vault(secrets) if secrets else ''

        bind_config = {
            'entrypoint':    "fixuid -q",
            'bind_volumes':  volumes,
            'forward_x11':   options.get('forward_x11'),
            'dind':          options.get('docker_in_docker'),
            'hostname':      workspace.get('hostname'),
            'command':       workspace.get('default_shell'),
            'user':          workspace.get('user'),
            'vault':         str_vault,
            'version':       'v' + __version__,
        }

        if not self.image_manager.experimental:
            warning(
                'Docker is not in experimental mode, which is required to squash layers.')
            warning(
                'To significantly reduce image size, please consider enabling it.')

        with waiter('Building image'):
            labels = {'kitt-config': json.dumps(bind_config)}

            self.image_manager.build(
                'kitt', template, name, labels=labels, pull=True)

        success('Build success !')

    def list(self):
        """List local kitt images
        """
        self.image_manager.list(repository='kitt')

    def remove(self, name: str):
        """Remove local kitt image

        Args:
            name (str): image name
        """
        self.image_manager.remove('kitt', name)
        success('Done !')

    def prune(self):
        """Prune force all kitt images
        """
        with waiter('Removing local images'):
            self.image_manager.prune('kitt')

    def refresh(self):
        """Pull latest version of local kitt images
        """
        with waiter('Updating local images'):
            self.image_manager.refresh('kitt')

    def pull(self, repository: str, tag: str):
        """Pull remote kitt image from repository

        Args:
            repository (str): full repository url
            tag (str): remote image tag
        """

        if tag == 'latest':
            warning('Tag "latest" is deprecated.')
            warning('Use kitt image descriptor instead (Ex. "devops").')

        with waiter(f'Pulling image { tag } from registry'):
            self.image_manager.pull(repository, tag, 'kitt')

        self.image_manager.remove(repository, tag)

        labels = self.image_manager.labels('kitt', tag)
        if 'kitt-config' not in labels:
            warning('Image does not look like a kitt image')

        success('Pull done !')

    def push(self, repository: str, name: str):
        """Push local kitt image to remote repository

        Args:
            repository (str): full repository url
            name (str): local image name
        """
        if not self.image_manager.stat('kitt', name):
            panic(f'Image { name } not found')

        with waiter(f'Pushing image { name }'):
            self.image_manager.push(repository, 'kitt', name)

        success('Push done !')

    def inspect(self, name: str):
        """Show image runtime configuration metadata

        Args:
            name (str): image name
        """
        if not (config := self._config(name)):
            panic('Could not load image config metadata')

        config = json.dumps(config, indent=4)

        info(config)

    def patch(self, name: str):
        """Quick patch runtime kitt configuration (avoid rebuild)

        Args:
            name (str): image name
        """
        if not (config := self._config(name)):
            panic('Could not load image config metadata')

        config = json.dumps(config, indent=4)

        fd, fname = tempfile.mkstemp()
        with open(fd, 'w', encoding='utf-8') as tmpfile:
            tmpfile.write(config)

        editor = os.environ.get('EDITOR', 'vi')
        subprocess.call(
            args=f'{ editor } { fname }',
            shell=True
        )

        with open(fname, 'r', encoding='utf-8') as tmpfile:
            config = json.dumps(json.load(tmpfile))

        # !! Make sure kitt-vault is preserved
        # !! Image dockerfile better generation
        with waiter(f'Building patched image "{ name }-patch"'):
            self.image_manager.build(
                name='kitt',
                template=f'FROM kitt:{ name }',
                tag=f'{ name }-patch',
                labels={'kitt-config': config},
                squash=False,
            )

        os.unlink(fname)
        success('Patch success !')

    def _config(self, name: str) -> dict:
        """Kitt image config metadata

        Args:
            name (str): kitt image

        Returns:
            dict: JSON loaded config
        """

        labels = self.image_manager.labels('kitt', name)
        if 'kitt-config' not in labels:
            return None

        labels = labels.get('kitt-config')

        try:
            return json.loads(labels)
        except json.decoder.JSONDecodeError:
            return None


# Where should I put you ??
def unpack_volume(volume: str) -> Tuple[str, str, str]:
    """Unpack docker like volume request to exploitable configuration tuple.

    Args:
        volume (str): docker like volume

    Returns:
        Tuple[str, str, str]: host, bind, mode
    """

    chunks = volume.split(':')
    nchunks = len(chunks)
    if not 1 < nchunks < 4:
        panic('Invalid --volume flag format')

    if nchunks == 2:
        host, bind = chunks
        return host, bind, 'ro'

    host, bind, mode = chunks
    if mode not in ['ro', 'rw']:
        panic(f'Unknown mode "{ mode }"')

    return host, bind, mode


client = KittClient()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# =============================================================================
# Author : @senges
# Version : See __version__.py
# Description : A portable shell
# =============================================================================


import click

from kitt.__version__ import __version__
from kitt.client import client
from kitt import logger


@click.group()
@click.help_option('-h', '--help')
@click.option('--debug', '-d', is_flag=True, help='Debug mode')
def main(debug):
    """main command group"""

    logger.config(debug)


@main.command('version')
def _version():
    """Show version"""

    logger.info('Kitt v' + __version__)


@main.command('run')
@click.help_option('-h', '--help')
@click.option('-v', '--volume', is_flag=False, multiple=True, help='Additional volume in OCI format')
@click.option('-u', '--user', is_flag=False, help='Run as other host user')
@click.option('--dind', is_flag=True, help='Enable docker in docker')
@click.argument('name', type=click.STRING)
def _run(name, volume, user, dind):
    """Run kitt shell"""

    extras = {
        "volumes": volume,
        "run_as": user,
        "dind": dind,
    }

    client.run(name, extras)


@main.command('list')
@click.help_option('-h', '--help')
def _list():
    """List local images"""

    client.list()


@main.command('remove')
@click.help_option('-h', '--help')
@click.argument("name")
def _remove(name):
    """Remove local image"""

    client.remove(name)


@main.command('prune')
@click.help_option('-h', '--help')
def _prune():
    """Prune local images"""

    client.prune()


@main.command('refresh')
@click.help_option('-h', '--help')
def _refresh():
    """Pull latest version of local images"""

    client.refresh()


@main.command('build')
@click.help_option('-h', '--help')
@click.argument('config', type=click.STRING)
@click.argument('name', type=click.STRING)
def _build(config, name):
    """Build image from source config file"""

    client.build(name, config)


@main.command('pull')
@click.help_option('-h', '--help')
@click.argument('registry', type=click.STRING)
@click.argument('name', type=click.STRING)
def _pull(registry, name):
    """Pull image and exit"""

    client.pull(registry, name)


@main.command('push')
@click.help_option('-h', '--help')
@click.argument('registry', type=click.STRING)
@click.argument('name', type=click.STRING)
def _push(registry, name):
    """Push kitt image to registry"""

    client.push(registry, name)


@main.command('inspect')
@click.help_option('-h', '--help')
@click.argument('image', type=click.STRING)
def _inspect(image):
    """Show image metadata"""

    client.inspect(image)


@main.command('patch')
@click.help_option('-h', '--help')
@click.argument('image', type=click.STRING)
def _patch(image):
    """Patch image runtime metadata"""

    client.patch(image)


@main.command('completion')
@click.help_option('-h', '--help')
@click.argument('shell', type=click.Choice(['bash', 'zsh', 'fish']))
def _completion(shell):
    """Shell auto-completion"""

    print(f'# Kitt {shell} auto-completion')
    print(f'eval "$(_KITT_COMPLETE={shell}_source kitt)"')


if __name__ == '__main__':
    main(False)



"""Kitt plugins definition and managment"""

import jinja2

from kitt.config import ConfigUtils
from kitt.logger import panic

def compose(name: str, config: dict) -> str:
    """generic compose method

    Args:
        name (str): plugin name
        config (dict): plugin configuration

    Returns:
       str: dockerfile string block
    """

    templater = ConfigUtils.mkpath(f'static/plugins/{name}.j2')
    try:
        with open(templater, 'r', encoding='utf-8') as file:
            template = file.read()
    except OSError:
        panic(f'Cannot load pluging "{name}" template')

    try:
        template = jinja2.Template(template)
        return template.render(config)
    except jinja2.TemplateError:
        panic(f'Cannot render pluging "{name}" : missing or invalid config data provided')

    raise Exception



import os
import uuid

from fs.tempfs import TempFS, errors as TempFSErrors

from kitt.crypto import b64, b64d, cipher_dict, secure_prompt, uncipher_dict
from kitt.logger import warning, info

class VaultFS:
    """Vault tmp filesystem to mount secrets
    """
    bindfs = None

    def __init__(self):
        self.bindfs = TempFS()
        self.fs_root = self.bindfs.getsyspath('/')

    def load(self, files: [str]) -> dict:
        """Load files in tmpfs

        Args:
            files [str]: base64 encoded files

        Returns:
            dict: file mapping list
        """

        volumes = {}

        for file in files:
            fname = str(uuid.uuid4())
            fpath = os.path.join(self.fs_root, fname)

            with open(fpath, 'wb+') as fout:
                fout.write(b64d(file['file']))

            volumes[fpath] = {
                'bind': file['location'],
                'mode': 'rw',
            }
            
        return volumes

    def close(self):
        """Close temp filesystem
        """
        try:
            self.bindfs.close()
        except AttributeError:
            pass
        except TempFSErrors.OperationFailed:
            warning('Could not properly remove local tempfs.')
            warning('Sensible data might remain on disk.')

def create_vault(config: dict) -> str:
    """Create base64 encoded encrypted vault to attach
    """

    info('Remember secret strengh is proportional to password strengh.')
    info('Most of the time, a strong password is a long password.')

    password = secure_prompt()

    vault = {
        'files': [],
        'envs': {},
    }

    for file in config.get('files', []):
        src, dest = file['src'], file['dest']
        with open(src, 'rb') as raw_file:
            raw = raw_file.read()
            raw = b64(raw)
        vault['files'].append({
            'location': dest,
            'file': raw
        })

    for env in config.get('envs', []):
        name, value = env['name'], env['value']
        vault['envs'][name] = value

    return cipher_dict(password, vault)

def load_vault(str_vault: str) -> dict:
    """Load base64 encoded encrypted vault
    """

    password = secure_prompt()

    return uncipher_dict(password, str_vault)